#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
(c) Bosch Sensortec GmbH, Reutlingen, Germany
Open Source as per the BSD-3 Clause
"""
from coinespy.coines import *


MAJOR_VERSION = 1
MINOR_VERSION = 0
MICRO_VERSION = 5

__version__ = f'{MAJOR_VERSION}.{MINOR_VERSION}.{MICRO_VERSION}'
