RELEASE_VERSION="0.11.0-4"
DISTRIBUTION_FILE_DATE="20220325-1731"
TARGET_PLATFORM="win32"
TARGET_ARCH="x64"
TARGET_MACHINE="x86_64"
TARGET_BITS="64"
HOST_UNAME="Linux"
USER_ID="1000"
USER_NAME="ilg"
GROUP_ID="1000"
GROUP_NAME="ilg"
CONTAINER_RUN_AS_ROOT="y"
HOST_WORK_FOLDER_PATH="/home/ilg/Work/openocd-0.11.0-4"
CONTAINER_WORK_FOLDER_PATH="/Host/home/ilg/Work/openocd-0.11.0-4"
HOST_CACHE_FOLDER_PATH="/home/ilg/Work/cache"
CONTAINER_CACHE_FOLDER_PATH="/Host/home/ilg/Work/cache"
DEPLOY_FOLDER_NAME="deploy"
