/*
 * Copyright (c) 2021 Martin Storsjo
 *
 * This file is part of llvm-mingw.
 *
 * Permission to use, copy, modify, and/or distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

#include "throwcatch-lib.h"
#include <string.h>
#include <stdio.h>

int main(int argc, char* argv[]) {
    int ret = 0;
    for (int i = 0; i <= 3; i++) {
        int caught = 0;
        try {
            libFunc(i);
        } catch (FirstException &e) {
            fprintf(stderr, "caught FirstException\n");
            caught = 1;
            if (strcmp(e.what(), "first")) {
                fprintf(stderr, "unexpected what()\n");
                ret = 1;
            }
        } catch (SecondException &e) {
            fprintf(stderr, "caught SecondException\n");
            caught = 2;
            if (strcmp(e.what(), "second")) {
                fprintf(stderr, "unexpected what()\n");
                ret = 1;
            }
        } catch (std::exception &e) {
            fprintf(stderr, "caught std::exception\n");
            caught = 3;
        } catch (...) {
            fprintf(stderr, "caught something else\n");
            caught = -1;
        }
        if (caught != i) {
            fprintf(stderr, "caught unexpected exception!\n");
            ret = 1;
        }
    }
    if (ret)
        fprintf(stderr, "got errors\n");
    else
        fprintf(stderr, "all ok\n");
    return ret;
}
